/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"
#include <QtPlugin>
#include <QInputDialog>
#include "ItemMarkerPlugin.h"
#include "PluginServices.h"
#include "TreeItemMarker.h"
#include "DefaultMarkerLabel.h"
#include "SystemTree.h"
#include "CubeLocation.h"
#include "CubeMetric.h"
#include "CallTree.h"

/**
 * If the plugin doesn't load, start cube with -verbose to get detailed information.
 * If the error message "Plugin verification data mismatch" is printed, check if the
 * same compiler and Qt version have been used.
 */

using namespace cubepluginapi;
using namespace itemmarkerplugin;
using namespace std;

bool
ItemMarkerPlugin::cubeOpened( PluginServices* service )
{
    this->service = service;
    connect( service, SIGNAL( contextMenuIsShown( cubepluginapi::DisplayType, cubepluginapi::TreeItem* ) ),
             this, SLOT( contextMenuIsShown( cubepluginapi::DisplayType, cubepluginapi::TreeItem* ) ) );

    markerLabel = new DefaultMarkerLabel( "Tree item marker" );
    marker      = service->getTreeItemMarker( "Tree item marker", QList<QPixmap>(), false, markerLabel );


    relationLabel = new DefaultMarkerLabel( "Related tree items" );
    QList<QPixmap> icons;
    icons.append( QPixmap( ":/images/relation.png" ) );
    relationMarker = service->getTreeItemMarker( "Related tree items", icons, false, relationLabel );
    QList<QPixmap> icons2;
    icons2.append( QPixmap( ":/images/jump-arrow-small.png" ) );
    relatedTabLabel  = new DefaultMarkerLabel( "Related tab" );
    relatedTabMarker = service->getTreeItemMarker( "Related tab", icons2, false, relatedTabLabel );

    // mark corresponding mpi item, if cuda item is selected
    initRelationMarker();
    connect( service, SIGNAL( treeItemIsSelected( cubepluginapi::TreeItem* ) ),
             this, SLOT( showRelationMarker() ) );

    // mark call site items
    initCallsiteMarker();
    connect( service, SIGNAL( treeItemIsSelected( cubepluginapi::TreeItem* ) ),
             this, SLOT( showCallSiteMarker() ) );
    connect( service, SIGNAL( tabActivated( cubepluginapi::DisplayType ) ),
             this, SLOT( tabChanged( cubepluginapi::DisplayType ) ) );

    service->addSettingsHandler( this );

    return true; // initialisation is ok => plugin should be shown
}

void
ItemMarkerPlugin::cubeClosed()
{
    service->disconnect();

    creatorItems.clear();
    markedItems.clear();
    markedRelationItems.clear();

    callSiteItems.clear();
    markedCallSiteRelationItems.clear();
}

/** search for call site items with corresponding kernel execution region */
void
ItemMarkerPlugin::initCallsiteMarker()
{
    cubegui::CallTree* callTree = dynamic_cast<cubegui::CallTree*>( service->getActiveTree( CALL ) );
    if ( !callTree )
    {
        return;
    }

    // callside id -> item
    QHash<QString, TreeItem*> callMap;
    QList<TreeItem* >         items = service->getTreeItems( cubepluginapi::DEFAULTCALLTREE );
    items += service->getTreeItems( cubepluginapi::TASKTREE );

    for ( TreeItem* item : items )
    {
        int idx = item->getName().indexOf( "callsite id=" );
        if ( idx >= 0 )
        {
            QString   idString    = item->getName().mid( idx );
            TreeItem* relatedItem = callMap.value( idString );
            if ( relatedItem )
            {
                callSiteItems[ item ]        = relatedItem;
                callSiteItems[ relatedItem ] = item;
            }
            else
            {
                callMap[ idString ] = item;
            }
            //qDebug() << "found call site item " << item->getName().mid( idx ) << item->getName();
        }
    }
}

/*
 * Special case task tree:
 * If a kernel item in the task tree is selected
 * 1. the corresponding kernel item in the default call tree is marked
 * 2. the corresponding task item in the default call tree is selected
 * The selection in the call tree (2.) causes deletion of the relation marker (1.), so this function marks
 * the previous marker again.
 * Known problems: If switching directly back to task tree after (2.), the marker is deleted.
 */
void
ItemMarkerPlugin::tabChanged( cubepluginapi::DisplayType )
{
    // changed tab to default task tree
    if ( !previousRelatedTabItems.isEmpty() )
    {
        for ( TreeItem* item : previousRelatedTabItems )
        {
            service->removeMarker( item, relationMarker );
            service->removeMarker( item, relatedTabMarker );
            service->addMarker( item, relationMarker );
            service->addMarker( item, relatedTabMarker );
            markedCallSiteRelationItems.append( item );
        }
        service->updateTreeView( DEFAULTCALLTREE );
        previousRelatedTabItems.clear();
    }
}

/** mark tree items pairs with the same call side id (experimental Score-P feature, id is part of the function name) */
void
ItemMarkerPlugin::showCallSiteMarker()
{
    previousRelatedTabItems = relatedTabItems;
    relatedTabItems.clear();

    for ( TreeItem* item : markedCallSiteRelationItems )
    {
        service->removeMarker( item, relationMarker );
        service->removeMarker( item, relatedTabMarker );
    }
    markedCallSiteRelationItems.clear();
    for ( TreeItem* item : service->getSelections( CALL ) )
    {
        TreeItem* relatedItem = callSiteItems[ item ];
        if ( relatedItem && !markedCallSiteRelationItems.contains( relatedItem ) )
        {
            markedCallSiteRelationItems.append( item );
            service->addMarker( item, relationMarker );
            service->addMarker( relatedItem, relationMarker );
            if ( item->getTree() != relatedItem->getTree() )
            {
                relatedTabItems.append( relatedItem );
                service->addMarker( item, relatedTabMarker );
                service->addMarker( relatedItem, relatedTabMarker );
            }
            markedCallSiteRelationItems.append( relatedItem );
            relationLabel->setLabel( item, "related to " + relatedItem->getName() );
            relationLabel->setLabel( relatedItem, "related to " + item->getName() );
            relatedTabLabel->setLabel( item, "related item in " + relatedItem->getTree()->getLabel() );
            relatedTabLabel->setLabel( relatedItem, "related item in " + item->getTree()->getLabel() );
        }
    }
    if ( !markedCallSiteRelationItems.isEmpty() )
    {
        previousRelatedTabItems.clear();
    }

    service->updateTreeView( DEFAULTCALLTREE );
}

/** search for related MPI tree item for each accelerator item and save them to relatedItems */
void
ItemMarkerPlugin::initRelationMarker()
{
    QList<TreeItem*>          acceleratorItems;
    QHash<QString, TreeItem*> processMap;

    for ( TreeItem* item : service->getActiveTree( SYSTEM )->getItems() )
    {
        cube::LocationGroup* location = dynamic_cast<cube::LocationGroup*>( item->getCubeObject() );
        if ( location )
        {
            if ( location->get_type() == cube::CUBE_LOCATION_GROUP_TYPE_ACCELERATOR )
            {
                acceleratorItems.append( item );
            }
            else if ( location->get_type() == cube::CUBE_LOCATION_GROUP_TYPE_PROCESS )
            {
                processMap[ item->getName() ] = item;
            }
        }
    }
    // fill relatedItems: acceleratorItem -> MPI item
    for ( TreeItem* acceleratorItem : acceleratorItems )
    {
        cube::LocationGroup*               accelerator = dynamic_cast<cube::LocationGroup*>( acceleratorItem->getCubeObject() );
        std::map<std::string, std::string> attrs       = accelerator->get_attrs();
        if ( attrs.size() != 0 )
        {
            map<string, string>::const_iterator ai;
            for ( ai = attrs.begin(); ai != attrs.end(); ++ai )
            {
                QString str = ai->second.c_str();
                if ( !str.isEmpty() )
                {
                    TreeItem* creatorItem = processMap[ str ];
                    if ( creatorItem )
                    {
                        creatorItems[ acceleratorItem ] = creatorItem;
                        childItems[ creatorItem ].append( acceleratorItem );
                        for ( TreeItem* child : acceleratorItem->getChildren() )
                        {
                            creatorItems[ child ] = creatorItem;
                        }
                        relationLabel->setLabel( creatorItem, "creates " + acceleratorItem->getName() );
                        relationLabel->setLabel( acceleratorItem, "is created by " + creatorItem->getName() );
                    }
                }
            }
        }
    }
}

/** mark related MPI tree item, if an accelerator item is selected */
void
ItemMarkerPlugin::showRelationMarker()
{
    for ( TreeItem* item : markedRelationItems )
    {
        service->removeMarker( item, relationMarker );
    }
    markedRelationItems.clear();
    for ( TreeItem* item : service->getSelections( SYSTEM ) )
    {
        TreeItem* relatedItem = creatorItems[ item ];
        if ( relatedItem && !markedRelationItems.contains( relatedItem ) )
        {
            markedRelationItems.append( relatedItem );
            service->addMarker( relatedItem, relationMarker );
        }
        else
        {
            for ( TreeItem* relatedItem : childItems[ item ] )
            {
                if ( !markedRelationItems.contains( relatedItem ) )
                {
                    markedRelationItems.append( relatedItem );
                    service->addMarker( relatedItem, relationMarker );
                }
            }
        }
        if ( markedRelationItems.size() > 0 )
        {
            if ( !markedRelationItems.contains( item ) )
            {
                markedRelationItems.append( item );
                service->addMarker( item, relationMarker );
            }
        }
    }
    service->updateTreeView( SYSTEMTREE );
}

/**
 * @brief DemoPlugin::contextMenuIsShown is called, if the user clicks with right mouse button on a tree
 * item.
 */
void
ItemMarkerPlugin::contextMenuIsShown( DisplayType type, TreeItem* item )
{
    contextItem = item;

    if ( item )
    {
        if ( !markedItems.contains( item ) )
        {
            QAction* contextAction = service->addContextMenuItem( type, tr( "Mark this item" ) );
            connect( contextAction, SIGNAL( triggered() ), this, SLOT( setMarker() ) );
        }
        else
        {
            QAction* contextAction = service->addContextMenuItem( type, tr( "Remove marker" ) );
            connect( contextAction, SIGNAL( triggered() ), this, SLOT( removeMarker() ) );
            contextAction = service->addContextMenuItem( type, tr( "Set marker label" ) );
            connect( contextAction, SIGNAL( triggered() ), this, SLOT( editMarker() ) );
        }
    }
}

void
ItemMarkerPlugin::removeMarker()
{
    service->removeMarker( contextItem, marker );
    service->updateTreeView( contextItem->getDisplayType() );
    markedItems.removeAll( contextItem );
}

void
ItemMarkerPlugin::setMarker()
{
    service->addMarker( contextItem, marker );
    service->updateTreeView( contextItem->getDisplayType() );
    markedItems.append( contextItem );
}

void
ItemMarkerPlugin::editMarker()
{
    bool    ok;
    QString text = QInputDialog::getText( service->getParentWidget(), tr( "Set marker label" ),
                                          tr( "Insert label:" ), QLineEdit::Normal, markerLabel->getLabel( contextItem ), &ok );
    if ( ok && !text.isEmpty() )
    {
        markerLabel->setLabel( contextItem, text );
    }
}

/** set a version number, the plugin with the highest version number will be loaded */
void
ItemMarkerPlugin::version( int& major, int& minor, int& bugfix ) const
{
    major  = 1;
    minor  = 0;
    bugfix = 0;
}

/** unique plugin name */
QString
ItemMarkerPlugin::name() const
{
    return "TreeItem Marker";
}

QString
ItemMarkerPlugin::getHelpText() const
{
    return tr( "This Plugin marks related items in the call, task and system tree. It also allows the user to mark tree items manually using the context menu." );
}

void
ItemMarkerPlugin::loadExperimentSettings( QSettings& settings )
{
    QList<QVariant> list      = settings.value( "marked" ).toList();
    QList<QVariant> labelList = settings.value( "markedLabel" ).toList();

    for ( int i = 0; i < list.size(); i++ )
    {
        TreeItem* item  = TreeItem::convertQVariantToTreeItem( list.at( i ) );
        QString   label = labelList.at( i ).toString();
        if ( item )
        {
            contextItem = item;
            setMarker();
            markerLabel->setLabel( contextItem, label );
        }
    }
}

void
ItemMarkerPlugin::saveExperimentSettings( QSettings& settings )
{
    QList<QVariant> list;
    QStringList     labelList;
    foreach( TreeItem * item, markedItems )
    {
        QVariant data = item->convertToQVariant();
        list.append( data );
        labelList.append( markerLabel->getLabel( item ) );
    }
    settings.setValue( "marked", QVariant( list ) );
    settings.setValue( "markedLabel", QVariant( labelList ) );
}
